"""Module information."""
import os

__title__ = "beetl"
__description__ = """
    BeETL is a Python package for extracting data from one datasource, 
    transforming it and loading it into another datasource
"""

__version__ = os.getenv("GHRELEASE", "0.3.2")
if os.getenv('GHRUN', False) and os.getenv('GHBRANCH', 'develop') == 'develop':
    __version__ += f".{os.getenv('RUN_ID')}"

__author__ = "Lars Scheibling"
__author_email__ = "lars.scheibling@hoglandet.se"
__license__ = "GnuPG 3.0"

__url__ = f"https://github.com/Hoglandets-IT/{__title__}"



from typing import List
import polars as pl
from .interface import TransformerInterface, register_transformer_class


@register_transformer_class("structs")
class StructTransformers(TransformerInterface):
    @staticmethod
    def staticfield(data: pl.DataFrame, field: str, value):
        """Add a struct field to the DataFrame

        Args:
            data (pl.DataFrame): The dataFrame to modify
            field (str): The field to add
            value (str): The value of the field to add

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        data = data.with_columns(pl.Series(field, [value] * len(data)))
        return data



import polars as pl
import pandas as pd
import sqlalchemy as sqla
import pyodbc
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)


class SqlserverConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for SQLServer sources"""

    table: str = None
    query: str = None

    def __init__(self, columns: list, table: str = None, query: str = None):
        super().__init__(columns)
        self.table = table
        self.query = query


class SqlserverConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for SQLServer sources"""

    data: pl.DataFrame
    connection_string: str
    query: str = None
    table: str = None

    def __init__(self, settings: dict):
        if settings.get("connection_string", False):
            self.connection_string = settings["connection_string"]
            if "driver" not in self.connection_string:
                if len(pyodbc.drivers()) == 0:
                    raise pyodbc.DatabaseError(
                        "No ODBC drivers found for SQL Server/pyodbc. "
                        "Please make sure at least one is installed"
                    )
                print("Found other ODBDC driver than the one provided, trying...\n")
                for driver in pyodbc.drivers():
                    print(f"Trying driver: {driver}")
                    if "SQL Server" in driver:
                        if "?" not in self.connection_string:
                            self.connection_string += f"?driver={driver}"
                        else:
                            self.connection_string += f"&driver={driver}"
                        break

                if "driver" not in self.connection_string:
                    raise pyodbc.DatabaseError(
                        "No working ODBC drivers found for SQL Server/pyodbc. "
                        "Please make sure at least one is installed"
                    )

            return

        self.connection_string = "mssql://"
        f"{settings['username']}:{settings['password']}"
        f"@{settings['host']}:{settings['port']}/{settings['database']}"


@register_source("sqlserver", SqlserverConfiguration, SqlserverConnectionSettings)
class SqlserverSource(SourceInterface):
    ConnectionSettingsClass = SqlserverConnectionSettings
    SourceConfigClass = SqlserverConfiguration

    """ A source for SqlServer data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(
        self, params=None, customQuery: str = None, returnData: bool = True
    ) -> pl.DataFrame:
        query = customQuery

        if query is None:
            if self.source_configuration.query is not None:
                query = self.source_configuration.query

            if query is None:
                if self.source_configuration.table is None:
                    raise Exception("No query or table specified")

                cols = ",".join(
                    ["[" + col.name + "]" for col in self.source_configuration.columns]
                )

                query = f"SELECT {cols} FROM [{self.source_configuration.table}]"

        if returnData:
            pdd = pd.read_sql_query(
                sql=query, con=self.connection_settings.connection_string
            )
            # pdd = pd.read_sql_query(
            #     query=query,
            #     connection_uri=self.connection_settings.connection_string
            # )
            return pl.from_pandas(pdd)
            # return pl.read_sql(
            #     query=query,
            #     connection_uri=self.connection_settings.connection_string
            # )
        with sqla.create_engine(
            self.connection_settings.connection_string
        ).connect() as con:
            if isinstance(query, str):
                query = sqla.text(query)

            con.execute(query)
            con.commit()

    def _insert(
        self, data: pl.DataFrame, table: str = None, connection_string: str = None
    ):
        if table is None:
            table = self.source_configuration.table

        if connection_string is None:
            connection_string = self.connection_settings.connection_string

        try:
            data.write_database(
                table, self.connection_settings.connection_string, if_exists="append"
            )
        except ModuleNotFoundError:
            data.write_database(
                table,
                self.connection_settings.connection_string.replace(
                    "mysql://", "mysql+pymysql://"
                ),
                if_exists="append",
            )

        return len(data)

    def insert(self, data: pl.DataFrame):
        return self._insert(data)

    def update(self, data: pl.DataFrame):
        tempDB = "##" + self.source_configuration.table + "_udtemp"

        # Insert data into temporary table
        self._insert(data, table=tempDB)

        field_spec = ", ".join(
            (
                f"TDEST.[{fld.name}] = TTEMP.[{fld.name}]"
                for fld in self.source_configuration.columns
                if not fld.skip_update and not fld.unique
            )
        )

        on_clause = " AND ".join(
            (
                f"TDEST.[{fld.name}] = TTEMP.[{fld.name}]"
                for fld in self.source_configuration.columns
                if fld.unique
            )
        )

        query = f"""
            UPDATE TDEST
            SET {field_spec}
            FROM {self.source_configuration.table} AS TDEST
            INNER JOIN {tempDB} AS TTEMP ON {on_clause}
        """

        self._query(customQuery=query, returnData=False)
        # self._query(customQuery="DROP TABLE " + tempDB, returnData=False)

        return len(data)

    def delete(self, data: pl.DataFrame):
        tempDB = "##" + self.source_configuration.table + "_ddtemp"
        self._insert(data, table=tempDB)

        where_clause = " AND ".join(
            (
                f"[{fld.name}] IN (SELECT [{fld.name}] FROM {tempDB}) "
                for fld in self.source_configuration.columns
                if fld.unique
            )
        )

        where_clause = " AND ".join(
            (
                f"TTEMP.[{fld.name}] = TDEST.[{fld.name}]"
                for fld in self.source_configuration.columns
                if fld.unique
            )
        )

        query = f"""
            DELETE TDEST
            FROM {self.source_configuration.table} AS TDEST
            WHERE EXISTS (
                SELECT 1/0
                FROM {tempDB} AS TTEMP
                WHERE {where_clause}
            ) 
        """

        self._query(customQuery=query, returnData=False)

        return len(data)



import polars as pl
from enum import Enum
import sqlalchemy as sqla
from typing import List
from .interface import (
    register_source,
    SourceInterface,
    ColumnDefinition,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)


class RestSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for MySQL sources"""

    pass


class RestRequest:
    """Configuration representing a REST request"""

    # Request type, URL and Query Arguments
    # type: post/get/put/patch/delete
    rtype: str = None
    url: str
    qargs: dict = None

    # Headers
    headers: dict = None

    # Body types (only one should be set)
    # Replacement will occur on the insert, update, delete functions
    # according to the columns in the source
    # The syntax for this is {{colname}}
    json_body: dict = None
    post_body: dict = None
    plain_body: dict = None

    # Authentication Methods

    # Basic (str with user:pass)
    auth_basic: str = None

    # Bearer (str with prefix and token)
    auth_bearer: str = None


class RestSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for MySQL sources"""

    base_url: str
    query_data: RestRequest = None
    insert_data: RestRequest = None
    update_data: RestRequest = None
    delete_data: RestRequest = None

    connection_string: str
    query: str = None
    table: str = None

    def __init__(self, settings: dict):
        if settings.get("connection_string", False):
            self.connection_string = settings["connection_string"]
            return

        self.connection_string = "mysql+pymysql://"
        f"{settings['username']}:{settings['password']}"
        f"@{settings['host']}:{settings['port']}/{settings['database']}"


@register_source("mysql", RestSourceConfiguration, RestSourceConnectionSettings)
class RestSource(SourceInterface):
    ConnectionSettingsClass = RestSourceConnectionSettings
    SourceConfigClass = RestSourceConfiguration

    """ A source for MySQL data """

    def _configure(self):
        raise Exception("Not yet implemented")

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(
        self, params=None, customQuery: str = None, returnData: bool = True
    ) -> pl.DataFrame:
        pass

    def _insert(
        self, data: pl.DataFrame, table: str = None, connection_string: str = None
    ):
        raise NotImplementedError("Insert not implemented")

    def insert(self, data: pl.DataFrame):
        raise NotImplementedError("Insert not implemented")

    def update(self, data: pl.DataFrame):
        raise NotImplementedError("Update not implemented")

    def delete(self, data: pl.DataFrame):
        raise NotImplementedError("Delete not implemented")



from typing import List, Union
from .config import BeetlConfig, SyncConfiguration
from .transformers.interface import TransformerConfiguration
import polars as pl
from time import perf_counter

BENCHMARK = []


class Beetl:
    """The main class for BeETL. This class is responsible for orchestrating the ETL process."""

    config: BeetlConfig = None
    """Holds the BeETL Configuration"""

    def __init__(self, config: BeetlConfig):
        self.config = config

    @classmethod
    def from_yaml(cls, path: str, encoding: str = "utf-8") -> "Beetl":
        """Creates a Beetl instance from a YAML-formatted configuration file

        Args:
            path (str): The path to the file
            encoding (str, optional): The charset for the file. Defaults to 'utf-8'.

        Returns:
            Beetl: An instance of Beetl
        """
        return cls(BeetlConfig.from_yaml_file(path, encoding))

    @classmethod
    def from_json(cls, path: str, encoding: str = "utf-8") -> "Beetl":
        """Creates a Beetl instance from a JSON-formatted configuration file

        Args:
            path (str): The path to the file
            encoding (str, optional): The charset for the file. Defaults to 'utf-8'.

        Returns:
            Beetl: _description_
        """
        return cls(BeetlConfig.from_json_file(path, encoding))

    @staticmethod
    def compare_datasets(
        source: pl.DataFrame,
        destination: pl.DataFrame,
        keys: List[str] = ["id"],
        columns: List[str] = [],
    ) -> List[pl.DataFrame]:
        """
        This function uses polars DataFrames to quickly compare two datasets and return the differences.

        Polars is roughly 10-50x faster than Pandas for this task, but this can vary depending on the dataset.

        Args:
            source (pl.DataFrame): The source dataset
            destination (pl.DataFrame): The destination dataset
            keys (List[str], optional): Unique keys in the dataset. Defaults to ["id"].
            columns (List[str], optional): List of columns to compare for differences. Defaults to [].

        Returns:
            List[pl.DataFrame]: A list of Polars Dataframes (Insert, Update, Delete) containing the differences between source and destination
        """

        if isinstance(source, Union[list, set, tuple]):
            source = pl.DataFrame(source)

        if isinstance(destination, Union[list, set, tuple]):
            destination = pl.DataFrame(destination)

        # If source is empty, delete all in destination
        if len(source) == 0:
            return source, source, destination

        # If destination is empty, create all from source
        if len(destination) == 0:
            return source, destination, destination

        # Get all columns from destination if none are specified
        columns = destination.columns if len(columns) == 0 else columns

        try:
            # Get rows that only exist in source (Creates)
            create = source.join(destination, on=keys, how="anti")

            # Get rows that exist in both and have differing values (Updates)
            update = source.join(destination, on=keys, how="semi").join(
                destination, on=columns, how="anti"
            )

            # Get rows that only exist in destination (Deletes)
            delete = destination.join(source, on=keys, how="anti")

        except pl.ColumnNotFoundError as e:
            raise ValueError(
                "One or more comparison columns do not exist \n"
                f"Source columns: {','.join(source.columns)} \n"
                f"Destination columns: {','.join(destination.columns)} \n"
                f"Comparison columns: {','.join(columns)} \n"
            ) from e

        return (create, update.select(keys + columns), delete.select(keys))

    def benchmark(self, text: str) -> None:
        """Inserts a benchmark into the log"""
        BENCHMARK.append({"text": text, "perf": perf_counter()})

        if len(BENCHMARK) > 1:
            print(
                BENCHMARK[-1]["text"]
                + ": "
                + str(round(BENCHMARK[-1]["perf"] - BENCHMARK[-2]["perf"], 5))
            )

    def runTransformers(
        self,
        source: pl.DataFrame,
        transformers: List[TransformerConfiguration],
        sync: SyncConfiguration,
    ) -> pl.DataFrame:
        transformed = source.clone()

        if transformers is not None and len(transformers) > 0:
            for transformer in transformers:
                if transformer.include_sync:
                    transformed = transformer.transform(transformed, sync=sync)
                    continue

                transformed = transformer.transform(transformed)

        return transformed

    def sync(self) -> None:
        """Executes the ETL process. The following steps will be performed:

        1. Load source and destination data

        2. Format source data to be compatible with destination through field and source transformers

        3. Compare source and destination data with compare_datasets

        4. Execute the respective insert, update and delete queries

        """
        self.benchmark("Starting sync and retrieving source data")
        for sync in self.config.sync_list:
            source_data = sync.source.query(sync.sourceConfig)
            self.benchmark("Finished data retrieval from source")

            destination_data = sync.destination.query(sync.destinationConfig)
            self.benchmark("Finished data retrieval from destination")

            self.benchmark("Starting source data transformation")
            transformedSource = self.runTransformers(
                source_data, sync.sourceTransformers, sync
            )
            self.benchmark("Finished data transformation before comparison")

            self.benchmark("Starting comparison")
            create, update, delete = self.compare_datasets(
                transformedSource,
                destination_data,
                sync.destination.source_configuration.unique_columns,
                sync.destination.source_configuration.comparison_columns,
            )
            self.benchmark("Successfully extracted operations from dataset")

            amount = {}

            print(
                f"Insert: {len(create)}, Update: {len(update)}, Delete: {len(delete)}"
            )

            self.benchmark("Starting database operations")
            amount["inserts"] = sync.destination.insert(
                self.runTransformers(create, sync.insertionTransformers, sync)
            )

            self.benchmark("Finished inserts, starting updates")
            amount["updates"] = sync.destination.update(
                self.runTransformers(update, sync.insertionTransformers, sync)
            )

            self.benchmark("Finished updates, starting deletes")
            amount["deletes"] = sync.destination.delete(delete)

            self.benchmark("Finished deletes, sync finished")

            print("Inserted: " + str(amount["inserts"]))
            print("Updated: " + str(amount["updates"]))
            print("Deleted: " + str(amount["deletes"]))

        return amount



from polars import DataFrame as POLARS_DF
from .interface import (
    register_source,
    SourceInterface,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)


class MongoDBSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for MongoDB sources"""

    pass


class MongoDBSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for MongoDB sources"""

    pass


@register_source("mongodb", MongoDBSourceConfiguration, MongoDBSourceConnectionSettings)
class MongoDBSource(SourceInterface):
    ConnectionSettingsClass = MongoDBSourceConnectionSettings
    SourceConfigClass = MongoDBSourceConfiguration

    """ A source for MongoDB data """

    def _configure(self):
        raise Exception("Not yet implemented")

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> POLARS_DF:
        pass

    def insert(self, data: POLARS_DF):
        pass

    def update(self, data: POLARS_DF):
        pass

    def delete(self, data: POLARS_DF):
        pass



from typing import List
from polars import DataFrame as POLARS_DF
from .interface import (
    register_source,
    SourceInterface,
    ColumnDefinition,
    SourceInterfaceConfiguration,
    SourceInterfaceConnectionSettings,
)


class PostgresSourceConfiguration(SourceInterfaceConfiguration):
    """The configuration class used for Postgresql sources"""

    columns: List[ColumnDefinition]

    def __init__(self, columns: list):
        super().__init__(columns)


class PostgresSourceConnectionSettings(SourceInterfaceConnectionSettings):
    """The connection configuration class used for Postgresql sources"""

    data: POLARS_DF

    def __init__(self, settings: dict):
        pass


@register_source(
    "postgres", PostgresSourceConfiguration, PostgresSourceConnectionSettings
)
class PostgresSource(SourceInterface):
    ConnectionSettingsClass = PostgresSourceConnectionSettings
    SourceConfigClass = PostgresSourceConfiguration

    """ A source for Postgresql data """

    def _configure(self):
        raise Exception("Not yet implemented")

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> POLARS_DF:
        pass

    def insert(self, data: POLARS_DF):
        pass

    def update(self, data: POLARS_DF):
        pass

    def delete(self, data: POLARS_DF):
        pass



from typing import List
import polars as pl
from .interface import TransformerInterface, register_transformer_class


@register_transformer_class("frames")
class FrameTransformer(TransformerInterface):
    @staticmethod
    def rename_columns(data: pl.DataFrame, columns: List[dict]):
        """Rename columns in the dataset

        Args:
            data (pl.DataFrame): The dataset
            columns (List[dict]): A list of dicts (from: col1, to: col2)

        Returns:
            pl.DataFrame: DataFrame with renamed columns
        """
        for column in columns:
            data.rename(column["from"], column["to"])
        return data

    @staticmethod
    def copy_columns(data: pl.DataFrame, columns: List[dict]):
        """Copies a given column to another column

        Args:
            data (pl.DataFrame): The dataset
            columns (List[dict]): A list of dicts (from: col1, to: col2)

        Returns:
            pl.DataFrame: DataFrame with renamed columns
        """
        for column in columns:
            data[column["to"]] = data[column["from"]]

        return data

    @staticmethod
    def drop_columns(data: pl.DataFrame, columns: List[str]) -> pl.DataFrame:
        """Drop columns from a DataFrame

        Args:
            data (pl.DataFrame): Drop columns from this dataframe
            columns (List[str]): The columns to dr op

        Returns:
            pl.DataFrame: The dataframe without the columns
        """
        return data.drop(columns)



from .interface import (
    register_transformer_class,
    TransformerInterface,
)
import polars as pl
from hashlib import sha1


def concat_and_sha(separator: str = "-", *args):
    retn = separator.join((x.lower() for x in args if x is not None))
    return sha1(retn.encode("utf-8")).hexdigest()


def never_none(value):
    if value is None or value == "":
        return "-"
    return value


@register_transformer_class("itop")
class ItopTransformer(TransformerInterface):
    @staticmethod
    def orgcode(data: pl.DataFrame, inFields: list, outField: str, toplevel: str):
        """Takes a number of columns, joins them and hashes the result"""

        def make_code(st):
            concat_and_sha("-", toplevel, *st.values())

        new = data.with_columns(pl.struct(inFields).apply(make_code).alias(outField))

        return new

    @staticmethod
    def orgtree(
        data: pl.DataFrame,
        treeFields: list,
        toplevel: str,
        name_field: str = "name",
        path_field: str = "orgpath",
        code_field: str = "code",
        parent_field: str = "parent_code",
    ):
        """Takes a number of columns and joins then together as an organization tree"""
        output = [
            {
                "status": "active",
                name_field: toplevel,
                path_field: f"Top->{toplevel}",
                code_field: concat_and_sha("-", toplevel),
                parent_field: None,
            }
        ]

        for i in range(len(treeFields)):
            for lvl in data[treeFields[: i + 1]].unique().iter_rows():
                lvls = [toplevel] + [never_none(x) for x in lvl]

                output.append(
                    {
                        "status": "active",
                        name_field: never_none(lvls[-1]),
                        path_field: "Top->" + "->".join(lvls),
                        code_field: concat_and_sha("-", *lvls),
                        parent_field: concat_and_sha("-", *lvls[:-1]),
                    }
                )

        return pl.DataFrame(output)

    @staticmethod
    def relations(data: pl.DataFrame, **kwargs) -> pl.DataFrame:
        """Insert data into iTop

        Args:
            data (pl.DataFrame): The dataFrame to modify
            connection (SourceInterface): The connection to the source
            configuration (SourceInterfaceConfiguration): The configuration for the source

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        if not kwargs.get("sync", False):
            raise Exception(
                "Error: include_sync option is not set for transformer itop.relations"
            )

        if kwargs["sync"].destination.source_configuration.has_foreign:
            transformed = data.clone()
            for field in kwargs["sync"].destination.source_configuration.columns:
                fk_def = getattr(field, "custom_options", {})
                if fk_def is not None:
                    fk_def = fk_def.get("itop", None)
                    if fk_def is not None:
                        try:
                            query = (
                                f'SELECT {fk_def["target_class"]}'
                                + f' WHERE {fk_def["reconciliation_key"]} = '
                            )

                            transformed = transformed.with_columns(
                                transformed[fk_def["comparison_field"]]
                                .apply(
                                    lambda x: query + f"'{x}'"
                                    if x is not None and x != ""
                                    else None
                                )
                                .alias(field.name)
                            )
                        except KeyError as e:
                            raise Exception(
                                "Error: The key definition is not valid"
                            ) from e

            return transformed

        return data



import polars as pl
from .interface import (
    TransformerInterface,
    register_transformer,
    register_transformer_class,
)


@register_transformer_class("strings")
class StringTransformer(TransformerInterface):
    @staticmethod
    def staticfield(data: pl.DataFrame, field: str, value: str) -> pl.DataFrame:
        """Add a static field to the DataFrame

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            stripChar (str): The character to strip

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        data = data.with_columns(pl.Series(field, [value] * len(data)))
        return data

    @staticmethod
    def strip(data: pl.DataFrame, inField: str, stripChars: str) -> pl.DataFrame:
        """Strip all given characters from a column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to process
            stripChar (str): The character to strip

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, inField)

        data = data.with_columns(data[inField].str.strip(stripChars))
        return data

    @staticmethod
    def lowercase(data: pl.DataFrame, inField: str, outField: str) -> pl.DataFrame:
        """Transform all values in a column to lowercase and insert
            them into another (or the same) column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to take in
            outField (str): The field to put the result in

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        __class__._validate_fields(data.columns, inField)

        data = data.with_columns(data[inField].str.to_lowercase().alias(outField))
        return data

    @staticmethod
    def uppercase(data: pl.DataFrame, inField: str, outField: str) -> pl.DataFrame:
        """Transform all values in a column to uppercase
            and insert them into another (or the same) column

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to take in
            outField (str): The field to put the result in

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        data = data.with_columns(data[inField].str.to_uppercase().alias(outField))
        return data

    @staticmethod
    def join(
        data: pl.DataFrame, inFields: list, outField: str, separator: str = ""
    ) -> pl.DataFrame:
        """Join multiple columns together

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inFields (list): The fields to join
            outField (str): The field to put the result in
            separator (str, optional): The separator to use. Defaults to ''.

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        nCol = pl.concat_str(data[inFields], separator=separator).alias(outField)

        return data.with_columns(nCol.alias(outField))

    @staticmethod
    def split(
        data: pl.DataFrame, inField: str, outFields: list, separator: str = ""
    ) -> pl.DataFrame:
        """Split a column into multiple fields

        Args:
            data (pl.DataFrame): The dataFrame to modify
            inField (str): The field to split
            outField (str): The fields to put the result in (in order)
            separator (str, optional): The separator to use. Defaults to ''.

        Returns:
            pl.DataFrame: The resulting DataFrame
        """
        amount = len(outFields)
        data = data.with_columns(
            data[inField]
            .str.splitn(separator, amount)
            .struct.unnest()
            .rename({f"field_{i}": fld for i, fld in enumerate(outFields)})
        )

        return data

    @staticmethod
    def quote(data: pl.DataFrame, inField: str, outField: str = None, quote: str = "'"):
        """Quotes the given column values"""

        tmpData = StringTransformer.staticfield(data, "ccQuoteFld", quote)

        new = StringTransformer.join(
            tmpData, ["ccQuoteFld", inField, "ccQuoteFld"], outField, separator=""
        )

        return new
